Documentation
¶
Index ¶
- type Config
- type Geometry
- func (g *Geometry) Display() string
- func (g *Geometry) Hash() []byte
- func (g *Geometry) KEM() kem.Scheme
- func (g *Geometry) Marshal() ([]byte, error)
- func (g *Geometry) NIKE() nike.Scheme
- func (g *Geometry) Scheme() (nike.Scheme, kem.Scheme)
- func (g *Geometry) String() string
- func (g *Geometry) Validate() error
Constants ¶
This section is empty.
Variables ¶
This section is empty.
Functions ¶
This section is empty.
Types ¶
type Geometry ¶
type Geometry struct {
// PacketLength is the length of a packet.
PacketLength int
// NrHops is the number of hops, this indicates the size
// of the Sphinx packet header.
NrHops int
// HeaderLength is the length of the Sphinx packet header in bytes.
HeaderLength int
// RoutingInfoLength is the length of the routing info portion of the header.
RoutingInfoLength int
// PerHopRoutingInfoLength is the length of the per hop routing info.
PerHopRoutingInfoLength int
// SURBLength is the length of SURB.
SURBLength int
// SphinxPlaintextHeaderLength is the length of the plaintext header.
SphinxPlaintextHeaderLength int
// PayloadTagLength is the length of the payload tag.
PayloadTagLength int
// ForwardPayloadLength is the size of the payload.
ForwardPayloadLength int
// UserForwardPayloadLength is the size of the usable payload.
UserForwardPayloadLength int
// NextNodeHopLength is derived off the largest routing info
// block that we expect to encounter. Everything else just has a
// NextNodeHop + NodeDelay, or a Recipient, both cases which are
// shorter.
NextNodeHopLength int
// SPRPKeyMaterialLength is the length of the SPRP key.
SPRPKeyMaterialLength int
// NIKEName is the name of the NIKE scheme used by the mixnet's Sphinx packet.
// NIKEName and KEMName are mutually exclusive.
NIKEName string
// KEMName is the name of the KEM scheme used by the mixnet's Sphinx packet.
// NIKEName and KEMName are mutually exclusive.
KEMName string
}
Geometry describes the geometry of a Sphinx packet.
Click to show internal directories.
Click to hide internal directories.